(* Lash *)
(* ported from Satallax file: searchoption.ml *)

open Flags
open Error
open Syntax
open Log

type searchoption =
 (*** ProcessProp1(m) might make pattern clauses and otherwise expands any remaining translucent defns leaving ProcessProp2 to apply the real rule ***)
    ProcessProp1 of ftm
 (*** ProcessProp2(m) For rules requiring one formula in the head (most rules) ***)
  | ProcessProp2 of ftm
 (*** Mating(plit,nlit,pl,nl) where pl,nl and plit = lit (h pl) and nlit = lit ~(h nl) ***)
  | Mating of int * int * stp * ftm list * ftm list
  | MatingList of int * ftm list * stp * (int * ftm list) list * bool
 (*** Confront(nplit,nnlit,a,u,v,l,r) where u,v,l,r:a and nplit = lit (u = v) and nnlit is lit (l != r) ***)
  | Confront of int * int * int * ftm * ftm * ftm * ftm
  | ConfrontList of int * int * ftm * ftm * (int * ftm * ftm) list * bool
 (*** DefaultElt(a) - create a default element of type a ***)
  | DefaultElt of string
 (*** DefaultEltInst(a) - create a default element of type a ***)
  | DefaultEltInst of string
 (*** NewInst(a,m) - put m:a in the set of instantiations ***)
  | NewInst of int * ftm
  (*** EnumInst enumerate the nth term to use as instantiation ***)
  | EnumInst of stp * int
 (*** Filter - use Minisat to filter usable sets ***)
  | Filter of int

let max_searchoptions = ref None
let searchoptions_retrieved = ref 0

(*** For printing a searchoption - helpful for debugging. - Chad, Oct 2010 ***)
let searchoption_str s =
  match s with
  | ProcessProp1(m) -> ("ProcessProp1(" ^ (ftm_str m) ^ ")")
  | ProcessProp2(m) -> ("ProcessProp2(" ^ (ftm_str m) ^ ")")
  | MatingList(plit,nlit,_,pl,nl) -> "MatingList..."
  | Confront(plit,nlit,a,s,t,u,v) -> "Confront(" ^ (string_of_int plit) ^ "," ^ (string_of_int nlit) ^ ")"
  | ConfrontList(ty,lit,t1,t2,l,listpos) -> "ConfrontList..."
  | DefaultElt(a) -> "DefaultElt(" ^ a ^ ")"
  | DefaultEltInst(a) -> "DefaultEltInst(" ^ a ^ ")"
  | EnumInst(a,n) -> "EnumInst(" ^ (stp_str a) ^ "," ^ (string_of_int n) ^ ")"
  | NewInst(_,m) -> "NewInst(" ^ (ftm_str m) ^ ")"
  | Filter(d) -> "Filter(" ^ (string_of_int d) ^ ")"

let ftm_list_size = List.fold_left (fun s e -> s + Ftm.size e) 0;;
let ftm_list_max = List.fold_left (fun s e -> max s (Ftm.size e)) 0;;
let ilog i = int_of_float (log (float_of_int (1 + i)))

let searchoption_priority oldp = function
  | ProcessProp1(m) | ProcessProp2(m) -> ilog (Ftm.size m)
  | _ -> 5 + oldp;;

                (*
let adaptPriorityModels n dt df du o =
  match o with
  | ProcessProp1 p ->
      let (mt,mf) = Models.count_true_false_interps p (List.map (fun (frint,_) -> frint) (Models.current_interp_2s())) in
      if !verbosity > 5 then Printf.printf "Proposition %s true in %d models and false in %d models.\n" (trm_str p) mt mf;
      n + dt * mt + df * mf + if mt = 0 || mf = 0 then du else 0
  | _ -> n
                 *)

(* Vectors *)
type 'a vec = {mutable data: 'a array; item: 'a};;
let vector_make len item = {data=Array.make len item; item=item};;
let vector_get v i = v.data.(i);;
let vector_get_default v i d =
  if i < Array.length v.data then Array.unsafe_get v.data i else d;;
let vector_double v =
  let ndata = Array.make (2 * Array.length v.data) v.item in
  Array.blit v.data 0 ndata 0 (Array.length v.data);
  v.data <- ndata;;
let vector_set v i j =
  while (Array.length v.data <= i) do vector_double v done;
  Array.unsafe_set v.data i j;;
let vector_update v i f =
  while (Array.length v.data <= i) do vector_double v done;
  Array.unsafe_set v.data i (f (Array.unsafe_get v.data i));;
let vector_iter f v =
  for i = 0 to Array.length v.data - 1 do f i (Array.unsafe_get v.data i) done;;
let vector_clear v =
  for i = 0 to Array.length v.data - 1 do Array.unsafe_set v.data i v.item done;;

(* Priority Queues *)
external pq_push : int -> unit = "c_pq_push" [@@noalloc];;
external pq_top : unit -> int  = "c_pq_top"  [@@noalloc];;
external pq_pop : unit -> unit = "c_pq_pop"  [@@noalloc];;

(*** Version 0: Same priorities treated as stacks. Fairness ensured with an offset incremented every 2^logomodulus inserts. ***)
let pqueue_offset = ref 0;;

let (stack_data : (searchoption list) vec) = vector_make 256 [];;
let rec stack_reset () =
  let p = pq_top () in
  if p >= 0 then begin
      pq_pop ();
      vector_set stack_data p [];
      stack_reset ()
    end

let stack_insert logmodulus p o =
  let p = p + ((!pqueue_offset) asr logmodulus) in
  incr pqueue_offset;
  let upd_fun_s = function
    | [] -> pq_push p; [o]
    | v -> o :: v
  in
  vector_update stack_data p upd_fun_s

let stack_pop () =
  let p = pq_top () in
  if p = -1 then raise Not_found;
  p, match vector_get stack_data p with
     | [h] -> pq_pop (); vector_set stack_data p []; h
     | h :: t -> vector_set stack_data p t; h
     | [] -> assert false

let stack_peek () =
  let p = pq_top () in
  if p = -1 then raise Not_found else p

(*** Version 1: Same priorities treated as queues. Fairness ensured with an offset incremented every 2^logmodulus inserts. ***)
type l2 = L2 of searchoption list * searchoption list
let (queue_data : l2 vec) = vector_make 256 (L2 ([], []))

let rec queue_reset () =
  let p = pq_top () in
  if p >= 0 then begin
    pq_pop ();
    vector_set queue_data p (L2 ([], []));
    queue_reset ()
  end

let queue_insert logmodulus p o =
  let p = p + ((!pqueue_offset) asr logmodulus) in
  incr pqueue_offset;
  let upd_fun_q = function
    | L2 ([], []) -> pq_push p; L2 ([o], [])
    | L2 (l, r) -> L2 (l, o :: r)
  in
  vector_update queue_data p upd_fun_q

let queue_get () =
  let p = pq_top () in
  if p = -1 then raise Not_found;
  p, match vector_get queue_data p with
  | L2 ([h], []) | L2 ([], [h]) ->
     pq_pop (); vector_set queue_data p (L2 ([], [])); h
  | L2 (h :: t, r) ->
     vector_set queue_data p (L2 (t, r)); h
  | L2 ([], l) ->
     match List.rev l with
     | h :: t -> vector_set queue_data p (L2 (t, [])); h
     | _ -> assert false

let queue_peek () =
  let p = pq_top () in
  if p = -1 then raise Not_found else p;;

(* FIFO *)

let l1 = ref [] and l2 = ref [];;
let fifo_insert o = l2 := o :: !l2;;
let rec fifo_get () =
  match !l1 with
  | h :: t ->
     l1 := t; 0, h
  | [] ->
     match !l2 with
     | [] -> raise Not_found
     | _ -> l1 := List.rev !l2; l2 := []; fifo_get ();;
let fifo_peek () = match !l1, !l2 with [], [] -> raise Not_found | _ -> 0;;
let fifo_reset () = l1 := []; l2 := [];;

(* Dual-Queue *)

let h = Hashtbl.create 1000;;
let v = vector_make 1001 ([],[]);;
let nextin = ref 0;;
let nextout = ref 0;;
let pullord = ref [];;

let dualq_reset () = Hashtbl.clear h; vector_clear v; nextin := 0; nextout := 0; pullord := [];;

let rec mergein p = function
  | [] -> [p]
  | (h :: t) as l -> if p > h then h :: mergein p t else if p < h then p :: l else l;;

let dualqpush p o =
  Hashtbl.add h !nextin (p, o);
  let vupd = function
    | [],[] -> pullord := mergein p !pullord; ([!nextin],[])
    | l1,l2 -> l1, !nextin :: l2 in
  vector_update v p vupd;
  incr nextin;;

let counter = ref 0;;

let rec dualqvecpop () =
  if !nextout >= !nextin then raise Not_found;
  match Hashtbl.find h !nextout with
  | ret -> Hashtbl.remove h !nextout; incr nextout; ret
  | exception Not_found -> incr nextout; dualqvecpop ()

let rec dualqqueuepop = function
  | p :: pt ->
     let rec process_list = function
       | x :: xt, l2 ->
          vector_set v p (xt, l2);
          begin match Hashtbl.find h x with
          | ret -> Hashtbl.remove h x; ret
          | exception Not_found -> process_list (xt, l2)
          end
       | [], [] -> dualqqueuepop pt
       | [], l -> process_list (List.rev l, [])
     in
     process_list (vector_get_default v p ([], []))
  | [] -> raise Not_found;;


let dualqpop () =
  let ret =
    if !fiDUALQUEUE_MOD > 0 then (* Mostly by priority, rarely oldest *)
      if (!counter mod !fiDUALQUEUE_MOD) = 0 then dualqvecpop () else dualqqueuepop !pullord
    else
      if (!counter mod -(!fiDUALQUEUE_MOD)) = 0 then dualqqueuepop !pullord else dualqvecpop ()
  in
  incr counter; ret;;


(* Dual-Stack *)

let h = Hashtbl.create 1000;;
let v = vector_make 1001 [];;
let nextin = ref 0;;
let nextout = ref 0;;
let pullord = ref [];;

let duals_reset () = Hashtbl.clear h; vector_clear v; nextin := 0; nextout := 0; pullord := [];;

let rec mergein p = function
  | [] -> [p]
  | (h :: t) as l -> if p > h then h :: mergein p t else if p < h then p :: l else l;;

let dualspush p o =
  Hashtbl.add h !nextin (p, o);
  let vupd = function
    | [] -> pullord := mergein p !pullord; [!nextin]
    | l -> !nextin :: l in
  vector_update v p vupd;
  incr nextin;;

let counter = ref 0;;

let rec dualsvecpop () =
  if !nextout >= !nextin then raise Not_found;
  match Hashtbl.find h !nextout with
  | ret -> Hashtbl.remove h !nextout; incr nextout; ret
  | exception Not_found -> incr nextout; dualsvecpop ()

let rec dualstackpop = function
  | p :: pt ->
     let rec process_list = function
       | x :: xt ->
          vector_set v p xt;
          begin match Hashtbl.find h x with
          | ret -> Hashtbl.remove h x; ret
          | exception Not_found -> process_list xt
          end
       | [] -> dualstackpop pt
     in
     process_list (vector_get_default v p [])
  | [] -> raise Not_found;;

let dualspop () =
  let ret =
    if !fiDUALQUEUE_MOD > 0 then (* Mostly by priority, rarely oldest *)
      if (!counter mod !fiDUALQUEUE_MOD) = 0 then dualsvecpop () else dualstackpop !pullord
    else
      if (!counter mod -(!fiDUALQUEUE_MOD)) = 0 then dualstackpop !pullord else dualsvecpop ()
  in
  incr counter; ret;;

(* - *)

let insertWithPriority n o =
  match o with
  | ProcessProp1(m) when Ftm.processed_mem m -> ()
  | _ ->
     let n = if n < 0 then 0 else n in
     if (!verbosity > 8) then Printf.printf "Inserting option with priority %d: %s\n" n (searchoption_str o);
     match !fiPRIORITY_QUEUE_IMPL with
     | 0 -> stack_insert 10 n o
     | 1 -> queue_insert 10 n o
     | 2 -> dualqpush (searchoption_priority n o) o
     | 3 -> fifo_insert o
     | 10 -> dualqpush n o
     | 11 -> queue_insert 10 (searchoption_priority n o) o
     | 4 -> stack_insert  9 n o
     | 5 -> stack_insert  8 n o
     | 6 -> stack_insert  7 n o
     | 7 -> stack_insert  6 n o
     | 8 -> queue_insert  5 n o
     | 9 -> queue_insert 20 (searchoption_priority n o) o
     | 12 -> dualspush (searchoption_priority n o) o
     | _ -> assert false

let getNext () =
  let ret =
    match !fiPRIORITY_QUEUE_IMPL with
    | 1 | 11 | 8 | 9 -> queue_get ()
    | 2 | 10 -> dualqpop ()
    | 3 -> fifo_get ()
    | 0 | 4 | 5 | 6 | 7 -> stack_pop ()
    | 12 -> dualspop ()
    | n -> raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n)))
  in
  if (!verbosity > 49) then (Printf.printf "Got next option: %s\n" (searchoption_str (snd ret)));
  incr searchoptions_retrieved;
  (match !max_searchoptions with Some m when !searchoptions_retrieved >= m -> raise Timeout
  | _ -> ());
  ret

let peekAtNext () =
  match !fiPRIORITY_QUEUE_IMPL with
  | 1 | 11 | 8 | 9 -> queue_peek ()
  | 2 | 10 | 12 -> failwith "unimplemented"
  | 3 -> fifo_peek ()
  | 0 | 4 | 5 | 6 | 7 -> stack_peek ()
  | n -> raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n)))

let reset_pqueues () =
  match !fiPRIORITY_QUEUE_IMPL with
  | 1 | 8 | 9 | 11 -> queue_reset ()
  | 2 | 10 -> dualq_reset ()
  | 3 -> fifo_reset ()
  | 0 | 4 | 5 | 6 | 7 -> stack_reset ()
  | 12 -> duals_reset ()
  | _ -> assert false
